var jestMatcherUtils = require('jest-matcher-utils');
var lodash = require('lodash');

const MATCHER_SYMBOL = Symbol('matcher');
/**
 * Used to test if an expectation on an argument is a custom matcher.
 */

function isMatcher(f) {
  return !!(f && f[MATCHER_SYMBOL]);
}

/**
 * Special symbol denoting the call of a function.
 */
const ApplyProp = Symbol('apply');

const printProperty = property => {
  if (property === ApplyProp) {
    return '';
  }

  if (typeof property === 'symbol') {
    return `[${property.toString()}]`;
  }

  return `.${property}`;
};
const printArg = arg => // Call toJSON on matchers directly to avoid wrapping them in quotes.
isMatcher(arg) ? arg.toJSON() : jestMatcherUtils.printExpected(arg);
const printCall = (property, args) => {
  const prettyArgs = args.map(arg => printArg(arg)).join(', ');
  const prettyProperty = printProperty(property);
  return `${prettyProperty}(${prettyArgs})`;
};
const printReturns = ({
  isError,
  isPromise,
  value
}, min, max) => {
  let thenPrefix = '';

  if (isPromise) {
    if (isError) {
      thenPrefix += 'thenReject';
    } else {
      thenPrefix += 'thenResolve';
    }
  } else if (isError) {
    thenPrefix += 'thenThrow';
  } else {
    thenPrefix += 'thenReturn';
  }

  return `.${thenPrefix}(${jestMatcherUtils.printExpected(value)}).between(${min}, ${max})`;
};
const printWhen = (property, args) => {
  if (args) {
    return `when(() => ${jestMatcherUtils.EXPECTED_COLOR(`mock${printCall(property, args)}`)})`;
  }

  return `when(() => ${jestMatcherUtils.EXPECTED_COLOR(`mock${printProperty(property)}`)})`;
};
const printExpectation = (property, args, returnValue, min, max) => `${printWhen(property, args)}${printReturns(returnValue, min, max)}`;
const printRemainingExpectations = expectations => expectations.length ? `Remaining unmet expectations:
 - ${expectations.map(e => e.toJSON()).join('\n - ')}` : 'There are no remaining unmet expectations.';

class UnfinishedExpectation extends Error {
  constructor(pendingExpectation) {
    super(`There is an unfinished pending expectation:

${pendingExpectation.toJSON()}

Please finish it by setting a return value even if the value
is undefined.`);
  }

}
class MissingWhen extends Error {
  constructor() {
    super(`You tried setting a return value without an expectation.

Every call to set a return value must be preceded by an expectation.`);
  }

}
class UnexpectedAccess extends Error {
  constructor(property, expectations) {
    super(`Didn't expect ${jestMatcherUtils.EXPECTED_COLOR(`mock${printProperty(property)}`)} to be accessed.

If you expect this property to be accessed then please
set an expectation for it.

${printRemainingExpectations(expectations)}`);
  }

}
class UnexpectedCall extends Error {
  constructor(property, args, expectations) {
    super(`Didn't expect ${jestMatcherUtils.EXPECTED_COLOR(`mock${printCall(property, args)}`)} to be called.

${printRemainingExpectations(expectations)}`);
  }

}
class NotAMock extends Error {
  constructor() {
    super(`We couldn't find the mock.

Make sure you're passing in an actual mock.`);
  }

}
class UnmetExpectations extends Error {
  constructor(expectations) {
    super(`There are unmet expectations:

 - ${expectations.map(e => e.toJSON()).join('\n - ')}`);
  }

}
/**
 * Merge property accesses and method calls for the same property
 * into a single call.
 *
 * @example
 * mergeCalls({ foo: [{ arguments: undefined }, { arguments: [1, 2, 3] }] }
 * // returns { foo: [{ arguments: [1, 2, 3] } }
 */

const mergeCalls = callMap => new Map(Array.from(callMap.entries()).map(([property, calls]) => {
  const hasMethodCalls = calls.some(call => call.arguments);
  const hasPropertyAccesses = calls.some(call => !call.arguments);

  if (hasMethodCalls && hasPropertyAccesses) {
    return [property, calls.filter(call => call.arguments)];
  }

  return [property, calls];
}));

class UnexpectedCalls extends Error {
  constructor(unexpectedCalls, expectations) {
    const printedCalls = Array.from(mergeCalls(unexpectedCalls).entries()).map(([property, calls]) => calls.map(call => call.arguments ? jestMatcherUtils.EXPECTED_COLOR(`mock${printCall(property, call.arguments)}`) : jestMatcherUtils.EXPECTED_COLOR(`mock${printProperty(property)}`)).join('\n - ')).join('\n - ');
    super(`The following calls were unexpected:

 - ${printedCalls}

${printRemainingExpectations(expectations)}`);
  }

}
class NestedWhen extends Error {
  constructor(parentProp, childProp) {
    const snippet = `
const parentMock = mock<T1>();
const childMock = mock<T2>();

when(() => childMock${printProperty(childProp)}).thenReturn(...);
when(() => parentMock${printProperty(parentProp)}).thenReturn(childMock)
`;
    super(`Setting an expectation on a nested property is not supported.

You can return an object directly when the first property is accessed,
or you can even return a separate mock:
${snippet}`);
  }

}

exports.UnexpectedProperty = void 0;

(function (UnexpectedProperty) {
  /**
   * Throw an error immediately.
   *
   * @example
   * // Will throw "Didn't expect foo to be accessed".
   * const { foo } = service;
   *
   * // Will throw "Didn't expect foo to be accessed",
   * // without printing the arguments.
   * foo(42);
   */
  UnexpectedProperty[UnexpectedProperty["THROW"] = 0] = "THROW";
  /**
   * Return a function that will throw if called. This can be useful if your
   * code destructures a function but never calls it.
   *
   * It will also improve error messages for unexpected calls because arguments
   * will be captured instead of throwing immediately on the property access.
   *
   * The function will be returned even if the property is not supposed to be a
   * function. This could cause weird behavior at runtime, when your code expects
   * e.g. a number and gets a function instead.
   *
   * @example
   * // This will NOT throw.
   * const { foo } = service;
   *
   * // This will NOT throw, and might produce unexpected results.
   * foo > 0
   *
   * // Will throw "Didn't expect foo(42) to be called".
   * foo(42);
   */

  UnexpectedProperty[UnexpectedProperty["CALL_THROW"] = 1] = "CALL_THROW";
})(exports.UnexpectedProperty || (exports.UnexpectedProperty = {}));

/**
 * Unbox the expectation's return value.
 *
 * If the value is an error then throw it.
 *
 * If the value is a promise then resolve/reject it.
 */
const unboxReturnValue = ({
  isError,
  isPromise,
  value
}) => {
  if (isError) {
    if (value instanceof Error) {
      if (isPromise) {
        return Promise.reject(value);
      }

      throw value;
    }

    if (isPromise) {
      return Promise.reject(new Error(value));
    }

    throw new Error(value);
  }

  if (isPromise) {
    return Promise.resolve(value);
  }

  return value;
};

/**
 * An expectation repository with a configurable behavior for
 * unexpected property access.
 */

class FlexibleRepository {
  constructor(unexpectedProperty = exports.UnexpectedProperty.THROW) {
    this.unexpectedProperty = void 0;
    this.expectations = new Map();
    this.expectedCallStats = new Map();
    this.unexpectedCallStats = new Map();

    this.apply = args => this.get(ApplyProp)(...args);

    this.handlePropertyWithMatchingExpectations = (property, expectations) => {
      // Avoid recording call stats for function calls, since the property is an
      // internal detail.
      if (property !== ApplyProp) {
        // An unexpected call could still happen later, if the property returns a
        // function that will not match the given args.
        this.recordExpected(property, undefined);
      }

      const propertyExpectation = expectations.find(e => e.expectation.matches(undefined));

      if (propertyExpectation) {
        this.countAndConsume(propertyExpectation);
        return unboxReturnValue(propertyExpectation.expectation.returnValue);
      }

      return (...args) => {
        const callExpectation = expectations.find(e => e.expectation.matches(args));

        if (callExpectation) {
          this.recordExpected(property, args);
          this.countAndConsume(callExpectation);
          return unboxReturnValue(callExpectation.expectation.returnValue);
        }

        return this.getValueForUnexpectedCall(property, args);
      };
    };

    this.handlePropertyWithNoExpectations = property => {
      switch (property) {
        case 'toString':
          return () => 'mock';

        case '@@toStringTag':
        case Symbol.toStringTag:
        case 'name':
          return 'mock';
        // pretty-format

        case '$$typeof':
        case 'constructor':
        case '@@__IMMUTABLE_ITERABLE__@@':
        case '@@__IMMUTABLE_RECORD__@@':
          return null;

        case MATCHER_SYMBOL:
          return false;

        case ApplyProp:
          return (...args) => this.getValueForUnexpectedCall(property, args);

        default:
          return this.getValueForUnexpectedAccess(property);
      }
    };

    this.unexpectedProperty = unexpectedProperty;
  }

  add(expectation) {
    const {
      property
    } = expectation;
    const expectations = this.expectations.get(property) || [];
    this.expectations.set(property, [...expectations, {
      expectation,
      matchCount: 0
    }]);
  }

  clear() {
    this.expectations.clear();
    this.expectedCallStats.clear();
    this.unexpectedCallStats.clear();
  }

  // TODO: this returns any, but the interface returns unknown
  //   unknown causes errors in apply tests, and any causes bugs in bootstrapped SM
  get(property) {
    const expectations = this.expectations.get(property);

    if (expectations && expectations.length) {
      return this.handlePropertyWithMatchingExpectations(property, expectations);
    }

    return this.handlePropertyWithNoExpectations(property);
  }

  getAllProperties() {
    return Array.from(this.expectations.keys());
  }

  getCallStats() {
    return {
      expected: this.expectedCallStats,
      unexpected: this.unexpectedCallStats
    };
  }

  getUnmet() {
    return [].concat(...Array.from(this.expectations.values()).map(expectations => expectations.filter(e => e.expectation.min > e.matchCount).map(e => e.expectation)));
  }

  recordExpected(property, args) {
    const calls = this.expectedCallStats.get(property) || [];
    this.expectedCallStats.set(property, [...calls, {
      arguments: args
    }]);
  }

  recordUnexpected(property, args) {
    const calls = this.unexpectedCallStats.get(property) || [];
    this.unexpectedCallStats.set(property, [...calls, {
      arguments: args
    }]);
  }

  countAndConsume(expectation) {
    // eslint-disable-next-line no-param-reassign
    expectation.matchCount++;
    this.consumeExpectation(expectation);
  }

  consumeExpectation(expectation) {
    const {
      property,
      max
    } = expectation.expectation;
    const expectations = this.expectations.get(property);

    if (expectation.matchCount === max) {
      this.expectations.set(property, expectations.filter(e => e !== expectation));
    }
  }

  getValueForUnexpectedCall(property, args) {
    this.recordUnexpected(property, args);
    throw new UnexpectedCall(property, args, this.getUnmet());
  }

  getValueForUnexpectedAccess(property) {
    if (this.unexpectedProperty === exports.UnexpectedProperty.THROW) {
      this.recordUnexpected(property, undefined);
      throw new UnexpectedAccess(property, this.getUnmet());
    }

    return (...args) => {
      this.recordUnexpected(property, args);
      throw new UnexpectedCall(property, args, this.getUnmet());
    };
  }

}

/**
 * Matches a call with more parameters than expected because it is assumed the
 * compiler will check that those parameters are optional.
 *
 * @example
 * new StrongExpectation(
 *   'bar',
 *   deepEquals([1, 2, 3]),
 *   23
 * ).matches('bar', [1, 2, 3]) === true;
 */

class StrongExpectation {
  constructor(property, args, returnValue, exactParams = false) {
    this.property = void 0;
    this.args = void 0;
    this.returnValue = void 0;
    this.exactParams = void 0;
    this.matched = 0;
    this.min = 1;
    this.max = 1;
    this.property = property;
    this.args = args;
    this.returnValue = returnValue;
    this.exactParams = exactParams;
  }

  setInvocationCount(min, max = 1) {
    this.min = min;
    this.max = max;
  }

  matches(args) {
    if (!this.matchesArgs(args)) {
      return false;
    }

    this.matched++;
    return this.max === 0 || this.matched <= this.max;
  }

  isUnmet() {
    return this.matched < this.min;
  }

  matchesArgs(received) {
    if (this.args === undefined) {
      return !received;
    }

    if (!received) {
      return false;
    }

    if (this.exactParams) {
      if (this.args.length !== received.length) {
        return false;
      }
    }

    return this.args.every((arg, i) => arg.matches(received[i]));
  }

  toJSON() {
    return printExpectation(this.property, this.args, this.returnValue, this.min, this.max);
  }

}

class PendingExpectationWithFactory {
  constructor(createExpectation, concreteMatcher, exactParams) {
    this.createExpectation = void 0;
    this.concreteMatcher = void 0;
    this.exactParams = void 0;
    this.args = void 0;
    this.property = void 0;
    this.createExpectation = createExpectation;
    this.concreteMatcher = concreteMatcher;
    this.exactParams = exactParams;
  }

  setProperty(value) {
    if (this.property) {
      throw new UnfinishedExpectation(this);
    }

    this.property = value;
  }

  setArgs(value) {
    this.args = value;
  }

  finish(returnValue) {
    if (!this.property) {
      throw new MissingWhen();
    }

    const expectation = this.createExpectation(this.property, this.args, returnValue, this.concreteMatcher, this.exactParams);
    this.property = undefined;
    this.args = undefined;
    return expectation;
  }

  toJSON() {
    return printWhen(this.property, this.args);
  }

}

function _extends() {
  _extends = Object.assign || function (target) {
    for (var i = 1; i < arguments.length; i++) {
      var source = arguments[i];

      for (var key in source) {
        if (Object.prototype.hasOwnProperty.call(source, key)) {
          target[key] = source[key];
        }
      }
    }

    return target;
  };

  return _extends.apply(this, arguments);
}

/**
 * Match a custom predicate.
 *
 * @param cb Will receive the value and returns whether it matches.
 * @param toJSON An optional function that should return a string that will be
 *   used when the matcher needs to be printed in an error message. By default,
 *   it stringifies `cb`.
 *
 * @example
 * const fn = mock<(x: number) => number>();
 * when(() => fn(It.matches(x => x >= 0))).returns(42);
 *
 * fn(2) === 42
 * fn(-1) // throws
 */

const matches = (cb, {
  toJSON = () => `matches(${cb.toString()})`
} = {}) => {
  const matcher = {
    [MATCHER_SYMBOL]: true,
    matches: arg => cb(arg),
    toJSON
  };
  return matcher;
};

const removeUndefined = object => {
  if (Array.isArray(object)) {
    return object.map(x => removeUndefined(x));
  }

  if (!lodash.isObjectLike(object)) {
    return object;
  }

  return lodash.omitBy(object, lodash.isUndefined);
};
/**
 * Compare values using deep equality.
 *
 * @param expected
 * @param strict By default, this matcher will treat a missing key in an object
 *   and a key with the value `undefined` as not equal. It will also consider
 *   non `Object` instances with different constructors as not equal. Setting
 *   this to `false` will consider the objects in both cases as equal.
 *
 * @see It.is A matcher that uses strict equality.
 */


const deepEquals = (expected, {
  strict = true
} = {}) => matches(actual => {
  if (strict) {
    return lodash.isEqual(actual, expected);
  }

  return lodash.isEqual(removeUndefined(actual), removeUndefined(expected));
}, {
  toJSON: () => printArg(expected)
});
/**
 * Compare values using `Object.is`.
 *
 * @link https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Object/is
 *
 * @see It.deepEquals A matcher that uses deep equality.
 */


const is = expected => matches(actual => Object.is(actual, expected), {
  toJSON: () => `${jestMatcherUtils.printExpected(expected)}`
});
/**
 * Match any value, including `undefined` and `null`.
 *
 * @example
 * const fn = mock<(x: number, y: string) => number>();
 * when(() => fn(It.isAny(), It.isAny())).thenReturn(1);
 *
 * fn(23, 'foobar') === 1
 */


const isAny = () => matches(() => true, {
  toJSON: () => 'anything'
});
/**
 * Recursively match an object.
 *
 * Supports nested matcher.
 *
 * @param partial An optional subset of the expected objected.
 *
 * @example
 * const fn = mock<(foo: { x: number, y: number }) => number>();
 * when(() => fn(It.isObject({ x: 23 }))).returns(42);
 *
 * fn({ x: 100, y: 200 }) // throws
 * fn({ x: 23, y: 200 }) // returns 42
 *
 * @example
 * It.isObject({ foo: It.isString() })
 */


const isObject = partial => matches(actual => lodash.isMatchWith(actual, partial || {}, (argValue, partialValue) => {
  if (isMatcher(partialValue)) {
    return partialValue.matches(argValue);
  } // Let lodash handle it otherwise.


  return undefined;
}), {
  toJSON: () => partial ? `object(${jestMatcherUtils.printExpected(partial)})` : 'object'
});
/**
 * Match any number.
 *
 * @example
 * const fn = mock<(x: number) => number>();
 * when(() => fn(It.isNumber())).returns(42);
 *
 * fn(20.5) === 42
 * fn(NaN) // throws
 */


const isNumber = () => matches(actual => typeof actual === 'number' && !Number.isNaN(actual), {
  toJSON: () => 'number'
});
/**
 * Match a string, potentially by a pattern.
 *
 * @param matching The string has to match this RegExp.
 * @param containing The string has to contain this substring.
 *
 * @example
 * const fn = mock<(x: string, y: string) => number>();
 * when(() => fn(It.isString(), It.isString({ containing: 'bar' }))).returns(42);
 *
 * fn('foo', 'baz') // throws
 * fn('foo', 'bar') === 42
 */


const isString = ({
  matching,
  containing
} = {}) => {
  if (matching && containing) {
    throw new Error('You can only pass `matching` or `containing`, not both.');
  }

  return matches(actual => {
    var _matching$test;

    if (typeof actual !== 'string') {
      return false;
    }

    if (containing) {
      return actual.indexOf(containing) !== -1;
    }

    return (_matching$test = matching == null ? void 0 : matching.test(actual)) != null ? _matching$test : true;
  }, {
    toJSON: () => containing || matching ? `string(${jestMatcherUtils.printExpected(containing || matching)})` : 'string'
  });
};
/**
 * Match an array.
 *
 * Supports nested matchers.
 *
 * @param containing If given, the matched array has to contain ALL of these
 *   elements in ANY order.
 *
 * @example
 * const fn = mock<(arr: number[]) => number>();
 * when(() => fn(It.isArray())).thenReturn(1);
 * when(() => fn(It.isArray([2, 3]))).thenReturn(2);
 *
 * fn({ length: 1, 0: 42 }) // throws
 * fn([]) === 1
 * fn([3, 2, 1]) === 2
 *
 * @example
 * It.isArray([It.isString({ containing: 'foobar' })])
 */


const isArray = containing => matches(actual => {
  if (!Array.isArray(actual)) {
    return false;
  }

  if (!containing) {
    return true;
  }

  return containing.every(x => actual.find(y => {
    if (isMatcher(x)) {
      return x.matches(y);
    }

    return deepEquals(x).matches(y);
  }) !== undefined);
}, {
  toJSON: () => containing ? `array(${jestMatcherUtils.printExpected(containing)})` : 'array'
});
/**
 * Matches anything and stores the received value.
 *
 * This should not be needed for most cases, but can be useful if you need
 * access to a complex argument outside the expectation e.g. to test a
 * callback.
 *
 * @param name If given, this name will be printed in error messages.
 *
 * @example
 * const fn = mock<(cb: (value: number) => number) => void>();
 * const matcher = It.willCapture();
 * when(() => fn(matcher)).thenReturn();
 *
 * fn(x => x + 1);
 * matcher.value?.(3) === 4
 */


const willCapture = name => {
  let capturedValue;
  const matcher = {
    [MATCHER_SYMBOL]: true,
    matches: actual => {
      capturedValue = actual;
      return true;
    },
    toJSON: () => name != null ? name : 'captures',

    get value() {
      return capturedValue;
    }

  };
  return matcher;
};
/**
 * Contains argument matchers that can be used to ignore arguments in an
 * expectation or to match complex arguments.
 */


const It = {
  matches,
  deepEquals,
  is,
  isAny,
  isObject,
  isNumber,
  isString,
  isArray,
  willCapture
};

const defaults = {
  concreteMatcher: It.deepEquals,
  unexpectedProperty: exports.UnexpectedProperty.CALL_THROW,
  exactParams: false
};
let currentDefaults = defaults;
/**
 * Override strong-mock's defaults.
 *
 * @param newDefaults These will be applied to the library defaults. Multiple
 *   calls don't stack e.g. calling this with `{}` will clear any previously
 *   applied defaults.
 */

const setDefaults = newDefaults => {
  currentDefaults = _extends({}, defaults, newDefaults);
};

/**
 * Since `when` doesn't receive the mock subject (because we can't make it
 * consistently return it from `mock()`, `mock.foo` and `mock.bar()`) we need
 * to store a global state for the currently active mock.
 *
 * We also want to throw in the following case:
 *
 * ```
 * when(() => mock()) // forgot returns here
 * when(() => mock()) // should throw
 * ```
 *
 * For that reason we can't just store the currently active mock, but also
 * whether we finished the expectation or not.
 */

let activeMock;
const setActiveMock = mock => {
  activeMock = mock;
};
const getActiveMock = () => activeMock;
/**
 * Store a global map of all mocks created and their state.
 *
 * This is needed because we can't reliably pass the state between `when`
 * and `thenReturn`.
 */

const mockMap = new Map();
const getMockState = mock => {
  if (mockMap.has(mock)) {
    return mockMap.get(mock);
  }

  throw new NotAMock();
};
const setMockState = (mock, state) => {
  mockMap.set(mock, state);
};
const getAllMocks = () => Array.from(mockMap.entries());

const createProxy = traps => // eslint-disable-next-line no-empty-function
new Proxy(
/* istanbul ignore next */
() => {}, {
  get: (target, prop) => {
    if (prop === 'bind') {
      return (thisArg, ...args) => (...moreArgs) => traps.apply([...args, ...moreArgs]);
    }

    if (prop === 'apply') {
      return (thisArg, args) => traps.apply(args || []);
    }

    if (prop === 'call') {
      return (thisArg, ...args) => traps.apply(args);
    }

    return traps.property(prop);
  },
  apply: (target, thisArg, args) => traps.apply(args),
  ownKeys: () => traps.ownKeys(),

  getOwnPropertyDescriptor(target, prop) {
    const keys = traps.ownKeys();

    if (keys.includes(prop)) {
      return {
        configurable: true,
        enumerable: true
      };
    }

    return undefined;
  }

});

const createStub = (repo, pendingExpectation, getCurrentMode) => {
  const stub = createProxy({
    property: property => {
      if (getCurrentMode() === Mode.CALL) {
        return repo.get(property);
      }

      setActiveMock(stub);
      pendingExpectation.setProperty(property);
      return createProxy({
        property: childProp => {
          throw new NestedWhen(property, childProp);
        },
        apply: args => {
          pendingExpectation.setArgs(args);
        },
        ownKeys: () => {
          throw new Error('Spreading during an expectation is not supported.');
        }
      });
    },
    apply: args => {
      if (getCurrentMode() === Mode.CALL) {
        return repo.apply(args);
      }

      setActiveMock(stub);
      pendingExpectation.setProperty(ApplyProp);
      pendingExpectation.setArgs(args);
      return undefined;
    },
    ownKeys: () => {
      if (getCurrentMode() === Mode.CALL) {
        return repo.getAllProperties();
      }

      throw new Error('Spreading during an expectation is not supported.');
    }
  });
  return stub;
};

const strongExpectationFactory = (property, args, returnValue, concreteMatcher, exactParams) => new StrongExpectation(property, // Wrap every non-matcher in the default matcher.
args == null ? void 0 : args.map(arg => isMatcher(arg) ? arg : concreteMatcher(arg)), returnValue, exactParams);

var Mode;

(function (Mode) {
  Mode[Mode["EXPECT"] = 0] = "EXPECT";
  Mode[Mode["CALL"] = 1] = "CALL";
})(Mode || (Mode = {}));

let currentMode = Mode.CALL;
const setMode = mode => {
  currentMode = mode;
};

const getMode = () => currentMode;
/**
 * Create a type safe mock.
 *
 * @see {@link when} Set expectations on the mock using `when`.
 *
 * @param options Configure the options for this specific mock, overriding any
 *   defaults that were set with {@link setDefaults}.
 * @param options.unexpectedProperty Controls what happens when an unexpected
 *   property is accessed.
 * @param options.concreteMatcher The matcher that will be used when one isn't
 *   specified explicitly.
 * @param options.exactParams Controls whether the number of received arguments
 *   has to match the expectation.
 *
 * @example
 * const fn = mock<() => number>();
 *
 * when(() => fn()).thenReturn(23);
 *
 * fn() === 23;
 */


const mock = ({
  unexpectedProperty,
  concreteMatcher,
  exactParams
} = {}) => {
  const options = {
    unexpectedProperty: unexpectedProperty != null ? unexpectedProperty : currentDefaults.unexpectedProperty,
    concreteMatcher: concreteMatcher != null ? concreteMatcher : currentDefaults.concreteMatcher,
    exactParams: exactParams != null ? exactParams : currentDefaults.exactParams
  };
  const repository = new FlexibleRepository(options.unexpectedProperty);
  const pendingExpectation = new PendingExpectationWithFactory(strongExpectationFactory, options.concreteMatcher, options.exactParams);
  const stub = createStub(repository, pendingExpectation, getMode);
  setMockState(stub, {
    repository,
    pendingExpectation,
    options
  });
  return stub;
};

const createInvocationCount = expectation => ({
  between(min, max) {
    expectation.setInvocationCount(min, max);
  },

  /* istanbul ignore next */
  times(exact) {
    expectation.setInvocationCount(exact, exact);
  },

  /* istanbul ignore next */
  anyTimes() {
    expectation.setInvocationCount(0, 0);
  },

  /* istanbul ignore next */
  atLeast(min) {
    expectation.setInvocationCount(min, Infinity);
  },

  /* istanbul ignore next */
  atMost(max) {
    expectation.setInvocationCount(0, max);
  },

  /* istanbul ignore next */
  once() {
    expectation.setInvocationCount(1, 1);
  },

  /* istanbul ignore next */
  twice() {
    expectation.setInvocationCount(2, 2);
  }
  /* eslint-enable no-param-reassign, no-multi-assign */


});

const finishPendingExpectation = (returnValue, pendingExpectation, repo) => {
  const finishedExpectation = pendingExpectation.finish(returnValue);
  repo.add(finishedExpectation);
  return createInvocationCount(finishedExpectation);
};

const getError = errorOrMessage => {
  if (typeof errorOrMessage === 'string') {
    return new Error(errorOrMessage);
  }

  if (errorOrMessage instanceof Error) {
    return errorOrMessage;
  }

  return new Error();
};

const createReturns = (pendingExpectation, repository) => ({
  thenReturn: returnValue => finishPendingExpectation( // This will handle both thenReturn(23) and thenReturn(Promise.resolve(3)).
  {
    value: returnValue,
    isError: false,
    isPromise: false
  }, pendingExpectation, repository),
  thenThrow: errorOrMessage => finishPendingExpectation({
    value: getError(errorOrMessage),
    isError: true,
    isPromise: false
  }, pendingExpectation, repository),
  thenResolve: promiseValue => finishPendingExpectation({
    value: promiseValue,
    isError: false,
    isPromise: true
  }, pendingExpectation, repository),
  thenReject: errorOrMessage => finishPendingExpectation({
    value: getError(errorOrMessage),
    isError: true,
    isPromise: true
  }, pendingExpectation, repository)
});

/**
 * Set an expectation on a mock.
 *
 * The expectation must be finished by setting a return value, even if the value
 * is `undefined`.
 *
 * If a call happens that was not expected then the mock will throw an error.
 * By default, the call is expected to only be made once. Use the invocation
 * count helpers to expect a call multiple times.
 *
 * @param expectation A callback to set the expectation on your mock. The
 *   callback must return the value from the mock to properly infer types.
 *
 * @example
 * const fn = mock<() => void>();
 * when(() => fn()).thenReturn(undefined);
 *
 * @example
 * const fn = mock<() => number>();
 * when(() => fn()).thenReturn(42).atMost(3);
 *
 * @example
 * const fn = mock<(x: number) => Promise<number>();
 * when(() => fn(23)).thenResolve(42);
 */

const when = expectation => {
  setMode(Mode.EXPECT);
  expectation();
  setMode(Mode.CALL);
  const {
    pendingExpectation,
    repository
  } = getMockState(getActiveMock());
  return createReturns(pendingExpectation, repository);
};

/**
 * Remove any remaining expectations on the given mock.
 *
 * @example
 * const fn = mock<() => number>();
 *
 * when(() => fn()).thenReturn(23);
 *
 * reset(fn);
 *
 * fn(); // throws
 */

const reset = mock => {
  getMockState(mock).repository.clear();
};
/**
 * Reset all existing mocks.
 *
 * @see reset
 */

const resetAll = () => {
  getAllMocks().forEach(([mock]) => {
    reset(mock);
  });
};

const verifyRepo = repository => {
  const unmetExpectations = repository.getUnmet();

  if (unmetExpectations.length) {
    throw new UnmetExpectations(unmetExpectations);
  }

  const callStats = repository.getCallStats();

  if (callStats.unexpected.size) {
    throw new UnexpectedCalls(callStats.unexpected, unmetExpectations);
  }
};
/**
 * Verify that all expectations on the given mock have been met.
 *
 * @throws Will throw if there are remaining expectations that were set
 * using `when` and that weren't met.
 *
 * @throws Will throw if any unexpected calls happened. Normally those
 * calls throw on their own, but the error might be caught by the code
 * being tested.
 *
 * @example
 * const fn = mock<() => number>();
 *
 * when(() => fn()).thenReturn(23);
 *
 * verify(fn); // throws
 */

const verify = mock => {
  const {
    repository
  } = getMockState(mock);
  verifyRepo(repository);
};
/**
 * Verify all existing mocks.
 *
 * @see verify
 */

const verifyAll = () => {
  getAllMocks().forEach(([mock]) => {
    verify(mock);
  });
};

exports.It = It;
exports.mock = mock;
exports.reset = reset;
exports.resetAll = resetAll;
exports.setDefaults = setDefaults;
exports.verify = verify;
exports.verifyAll = verifyAll;
exports.when = when;
//# sourceMappingURL=index.js.map
