import type { ExpectationRepository } from '../expectation/repository/expectation-repository';
import type { PendingExpectation } from '../when/pending-expectation';
import type { InvocationCount } from './invocation-count';
export declare type PromiseStub<R, P> = {
    /**
     * Set the return value for the current call.
     *
     * @param value This needs to be of the same type as the value returned
     *   by the call inside `when`.
     *
     * @example
     * when(() => fn()).thenReturn(Promise.resolve(23));
     *
     * @example
     * when(() => fn()).thenReturn(Promise.reject({ foo: 'bar' });
     */
    thenReturn(value: P): InvocationCount;
    /**
     * Set the return value for the current call.
     *
     * @param promiseValue This needs to be of the same type as the value inside
     *   the promise returned by the `when` callback.
     */
    thenResolve(promiseValue: R): InvocationCount;
    /**
     * Make the current call reject with the given error.
     *
     * @param error An `Error` instance. If you want to reject with a non error
     *   then use the `thenReturn` method.
     */
    thenReject(error: Error): InvocationCount;
    /**
     * Make the current call reject with an error with the given message.
     *
     * @param message Will be wrapped in `new Error()`. If you want to reject
     *   with a custom error then pass it here instead of the message. If you
     *   want to reject with a non error then use `thenReturn`.
     */
    thenReject(message: string): InvocationCount;
    /**
     * Make the current call reject with `new Error()`.
     */
    thenReject(): InvocationCount;
};
export declare type NonPromiseStub<R> = {
    /**
     * Set the return value for the current call.
     *
     * @param returnValue This needs to be of the same type as the value returned
     *   by the `when` callback.
     */
    thenReturn(returnValue: R): InvocationCount;
    /**
     * Make the current call throw the given error.
     *
     * @param error The error instance. If you want to throw a simple `Error`
     *   you can pass just the message.
     */
    thenThrow(error: Error): InvocationCount;
    /**
     * Make the current call throw an error with the given message.
     *
     * @param message Will be wrapped in `new Error()`. If you want to throw
     *   a custom error pass it here instead of the message.
     */
    thenThrow(message: string): InvocationCount;
    /**
     * Make the current call throw `new Error()`.
     */
    thenThrow(): InvocationCount;
};
export declare const createReturns: (pendingExpectation: PendingExpectation, repository: ExpectationRepository) => {
    thenReturn: (returnValue: any) => InvocationCount;
    thenThrow: (errorOrMessage?: Error | string) => InvocationCount;
    thenResolve: (promiseValue: any) => InvocationCount;
    thenReject: (errorOrMessage?: Error | string) => InvocationCount;
};
